{% load static i18n %}

<div style="position: {{ preloader.position }};{{ preloader.style }}" {% if preloader.id %} id="{{ preloader.id }}"{% endif %}
  class="preloader {% if preloader.framed %}framed{% endif %} {% if preloader.backdrop %}backdrop{% endif %} {% if preloader.show %}show{% endif %} {% if not preloader.large %}small{% endif %}">
  <div class="contents">
    {% block contents %}
    <h1 class="text">
      {% translate "Loading, please wait" %}
      <span class="ellipsis-container">
        <span class="ellipsis-bg">...</span>
        <span class="ellipsis">...</span>
      </span>
    </h1>
    {% endblock contents %}
  </div>
</div>
