{% load i18n klaatu %}

<div class="modal dynamic-container {{ modal.classes }}" id="{{ modal.id }}" role="dialog" tabindex="-1"
  data-content-url="{{ modal.url }}"
  data-required-params="{{ modal.required_params }}"
  data-optional-params="{{ modal.optional_params }}"
  {% for param in modal.all_params %} data-{{ param }}="{% map_to_context param %}"{% endfor %}
  {% for key, value in modal.data_attrs.items %} {{ key }}="{{ value }}"{% endfor %}>
  <div class="modal-dialog {% if modal.scrollable %}modal-dialog-scrollable{% endif %} {% if modal.large %}modal-lg{% endif %} {% if modal.center %}modal-dialog-centered{% endif %}">
    <div class="w-100 position-relative">
      <div class="modal-content dynamic-content"></div>
      {% preloader large=modal.large %}
    </div>
  </div>
</div>
