{% load klaatu %}

{% block outer %}
<div class="modal-header">
  {% block header %}
  <h5 class="modal-title">{% block title %}{% endblock title %}</h5>
  <button type="button" data-bs-dismiss="modal" class="btn-close modal-header-button"></button>
  {% endblock header %}
</div>
<div class="modal-body">
  <div class="alerts">
    {% block alerts %}{% endblock alerts %}
  </div>
  {% block body %}{% endblock body %}
</div>
<div class="modal-footer">{% block footer %}{% endblock footer %}</div>
{% endblock outer %}
